@@ -18,6 +18,8 @@ import { CodeMirrorEditor, ICodeMirror } from '@jupyterlab/codemirror';
 
 import CodeMirror from 'codemirror';
 
+import { requestAPI } from './handler';
+
 import '../style/index.css';
 import spellcheckSvg from '../style/icons/ic-baseline-spellcheck.svg';
 
@@ -55,80 +57,92 @@ interface IContext {
   position: CodeMirror.Position;
 }
 
-interface ILanguage {
+/**
+ * Dictionary data defined by a pair of Hunspell .aff and .dic files.
+ * More than one dictionary may exist for a given language.
+ */
+interface IDictionary {
+  /**
+   * Identifier of the dictionary consisting of the filename without the .aff/.dic suffix
+   * and path information if needed to distinguish from other dictionaries.
+   */
+  id: string;
+  /**
+   * BCP 47 code identifier.
+   */
   code: string;
+  /**
+   * Display name, usually in the form "Language (Region)".
+   */
   name: string;
+  /**
+   * Path to the .aff file.
+   */
   aff: string;
+  /**
+   * Path to the .dic file.
+   */
   dic: string;
 }
 
-// English dictionaries come from https://github.com/en-wl/wordlist
-import en_aff from 'file-loader!../dictionaries/en_US.aff';
-import en_dic from 'file-loader!../dictionaries/en_US.dic';
+interface ILanguageManagerResponse {
+  version: string;
+  dictionaries: IDictionary[];
+}
 
-import en_gb_aff from 'file-loader!../dictionaries/en_GB-ise.aff';
-import en_gb_dic from 'file-loader!../dictionaries/en_GB-ise.dic';
+class LanguageManager {
+  languages: IDictionary[];
 
-import en_ca_aff from 'file-loader!../dictionaries/en_CA.aff';
-import en_ca_dic from 'file-loader!../dictionaries/en_CA.dic';
+  public ready: Promise<any>;
 
-import en_au_aff from 'file-loader!../dictionaries/en_AU.aff';
-import en_au_dic from 'file-loader!../dictionaries/en_AU.dic';
-
-import de_de_aff from 'file-loader!../dictionaries/de_DE_frami.aff';
-import de_de_dic from 'file-loader!../dictionaries/de_DE_frami.dic';
-
-import de_at_aff from 'file-loader!../dictionaries/de_AT_frami.aff';
-import de_at_dic from 'file-loader!../dictionaries/de_AT_frami.dic';
-
-import de_ch_aff from 'file-loader!../dictionaries/de_CH_frami.aff';
-import de_ch_dic from 'file-loader!../dictionaries/de_CH_frami.dic';
-
-import fr_fr_aff from 'file-loader!../dictionaries/fr.aff';
-import fr_fr_dic from 'file-loader!../dictionaries/fr.dic';
-
-import es_es_aff from 'file-loader!../dictionaries/es_ES.aff';
-import es_es_dic from 'file-loader!../dictionaries/es_ES.dic';
-
-//import it_it_aff from 'file-loader!../dictionaries/it_IT.aff';
-//import it_it_dic from 'file-loader!../dictionaries/it_IT.dic';
-
-import pt_pt_aff from 'file-loader!../dictionaries/pt_PT.aff';
-import pt_pt_dic from 'file-loader!../dictionaries/pt_PT.dic';
-
-const languages: ILanguage[] = [
-  { code: 'en-us', name: 'English (American)', aff: en_aff, dic: en_dic },
-  { code: 'en-gb', name: 'English (British)', aff: en_gb_aff, dic: en_gb_dic },
-  { code: 'en-ca', name: 'English (Canadian)', aff: en_ca_aff, dic: en_ca_dic },
-  {
-    code: 'en-au',
-    name: 'English (Australian)',
-    aff: en_au_aff,
-    dic: en_au_dic
-  },
-  {
-    code: 'de-de',
-    name: 'Deutsch (Deutschland)',
-    aff: de_de_aff,
-    dic: de_de_dic
-  },
-  {
-    code: 'de-at',
-    name: 'Deutsch (Österreich)',
-    aff: de_at_aff,
-    dic: de_at_dic
-  },
-  { code: 'de-ch', name: 'Deutsch (Schweiz)', aff: de_ch_aff, dic: de_ch_dic },
-  { code: 'fr-fr', name: 'Français (France)', aff: fr_fr_aff, dic: fr_fr_dic },
-  { code: 'es-es', name: 'Español (España)', aff: es_es_aff, dic: es_es_dic },
-  //{code: 'it-it', name: 'Italiano (Italia)', aff: it_it_aff, dic: it_it_dic},
-  {
-    code: 'pt-pt',
-    name: 'Português (Portugal)',
-    aff: pt_pt_aff,
-    dic: pt_pt_dic
+  /**
+   * initialise the manager
+   * mainly reading the definitions from the external extension
+   */
+  constructor() {
+    // read the list of languages from the external extension
+    this.ready = requestAPI<any>('language_manager').then(
+      (values: ILanguageManagerResponse) => {
+        console.debug('LanguageManager is ready');
+        this.languages = values.dictionaries;
+      }
+    );
   }
-];
+
+  /**
+   * get an array of languages, put "language" in front of the list
+   * the list is alphabetically sorted
+   */
+  getChoices(language: IDictionary) {
+    return [
+      language,
+      ...this.languages
+        .filter(l => l.id !== language.id)
+        .sort((a, b) => a.name.localeCompare(b.name))
+    ];
+  }
+
+  /**
+   * select the language by the identifier
+   */
+  getLanguageByIdentifier(identifier: string): IDictionary | undefined {
+    const exactMatch = this.languages.find(l => l.id === identifier);
+    if (exactMatch) {
+      return exactMatch;
+    }
+    // approximate matches support transition from the 0.5 version (and older)
+    // that used incorrect codes as language identifiers
+    const approximateMatch = this.languages.find(
+      l => l.id.toLowerCase() === identifier.replace('-', '_').toLowerCase()
+    );
+    if (approximateMatch) {
+      console.warn(
+        `Language identifier ${identifier} has a non-exact match, please update it to ${approximateMatch.id}`
+      );
+      return approximateMatch;
+    }
+  }
+}
 
 class StatusWidget extends ReactWidget {
   language_source: () => string;
@@ -150,10 +164,12 @@ class SpellChecker {
   dictionary: any;
   suggestions_menu: Menu;
   status_widget: StatusWidget;
+  status_msg = 'Dictionary not loaded';
 
   // Default Options
   check_spelling = true;
-  language: ILanguage;
+  language: IDictionary;
+  language_manager: LanguageManager;
   rx_word_char = /[^-[\]{}():/!;&@$£%§<>"*+=?.,~\\^|_`#±\s\t]/;
   rx_non_word_char = /[-[\]{}():/!;&@$£%§<>"*+=?.,~\\^|_`#±\s\t]/;
   ignored_tokens: Set<string> = new Set();
@@ -169,8 +185,8 @@ class SpellChecker {
     protected palette?: ICommandPalette,
     protected status_bar?: IStatusBar
   ) {
-    // have at least a default
-    this.language = languages[0];
+    // use the language_manager
+    this.language_manager = new LanguageManager();
 
     // read the settings
     this.setup_settings();
@@ -201,7 +217,11 @@ class SpellChecker {
   // move the load_dictionary into the setup routine, because then
   // we know that the values are set correctly!
   setup_settings() {
-    Promise.all([this.setting_registry.load(extension.id), this.app.restored])
+    Promise.all([
+      this.setting_registry.load(extension.id),
+      this.app.restored,
+      this.language_manager.ready
+    ])
       .then(([settings]) => {
         this.update_settings(settings);
         settings.changed.connect(() => {
@@ -226,10 +246,11 @@ class SpellChecker {
     this._set_theme(theme);
 
     // read the saved language setting
-    const language_code = settings.get('language').composite;
-    const user_language = languages.filter(l => l.code === language_code)[0];
+    const language_id = settings.get('language').composite as string;
+    const user_language =
+      this.language_manager.getLanguageByIdentifier(language_id);
     if (user_language === undefined) {
-      console.warn('The language ' + language_code + ' is not supported!');
+      console.warn('The language ' + language_id + ' is not supported!');
     } else {
       this.language = user_language;
       // load the dictionary
@@ -474,13 +495,16 @@ class SpellChecker {
   }
 
   load_dictionary() {
+    this.status_msg = 'Loading dictionary ...';
+    this.status_widget.update();
     return Promise.all([
       fetch(this.language.aff).then(res => res.text()),
       fetch(this.language.dic).then(res => res.text())
     ]).then(values => {
       this.dictionary = new Typo(this.language.name, values[0], values[1]);
-      console.log('Dictionary Loaded ', this.language.name, this.language.code);
+      console.debug('Dictionary Loaded ', this.language.name, this.language.id);
 
+      this.status_msg = this.language.name;
       // update the complete UI
       this.status_widget.update();
       this.refresh_state();
@@ -534,26 +558,38 @@ class SpellChecker {
   }
 
   choose_language() {
-    // show the current language first, then all others
-    const choices = [
-      this.language,
-      ...languages.filter(l => l.name !== this.language.name)
-    ];
+    const choices = this.language_manager.getChoices(this.language);
+
+    const choiceStrings = choices.map(
+      // note: two dictionaries may exist for a language with the same name,
+      // so we append the actual id of the dictionary in the square brackets.
+      dictionary => dictionary.name + ' [' + dictionary.id + ']'
+    );
 
     InputDialog.getItem({
       title: 'Choose spellchecker language',
-      items: choices.map(language => language.name)
+      items: choiceStrings
     }).then(value => {
       if (value.value !== null) {
-        this.language = languages.filter(l => l.name === value.value)[0];
-        // the setup routine will load the dirctionary
-        this.settings.set('language', this.language.code).catch(console.warn);
+        const index = choiceStrings.indexOf(value.value);
+        const lang = this.language_manager.getLanguageByIdentifier(
+          choices[index].id
+        );
+        if (!lang) {
+          console.error(
+            'Language could not be matched - please report this as an issue'
+          );
+          return;
+        }
+        this.language = lang;
+        // the setup routine will load the dictionary
+        this.settings.set('language', this.language.id).catch(console.warn);
       }
     });
   }
 
   setup_language_picker() {
-    this.status_widget = new StatusWidget(() => this.language.name);
+    this.status_widget = new StatusWidget(() => this.status_msg);
     this.status_widget.node.onclick = () => {
       this.choose_language();
     };
